@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2019, 2021,
+ * Copyright (C) 1986, 1988, 1989, 1991-2019, 2021, 2022,
  * the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
@@ -25,10 +25,8 @@
  */
 
 #include "awk.h"
+#include <math.h>
 
-extern double pow(double x, double y);
-extern double modf(double x, double *yp);
-extern double fmod(double x, double y);
 NODE **fcall_list = NULL;
 long fcall_count = 0;
 int currule = 0;
@@ -41,6 +39,8 @@ static int num_exec_hook = 0;
 static Func_pre_exec pre_execute[MAX_EXEC_HOOKS];
 static Func_post_exec post_execute = NULL;
 
+static double fix_nan_sign(double left, double right, double result);
+
 extern void frame_popped();
 
 int OFSlen;
@@ -240,6 +240,7 @@ static const char *const nodetypes[] = {
 	"Node_var",
 	"Node_var_array",
 	"Node_var_new",
+	"Node_elem_new",
 	"Node_param_list",
 	"Node_func",
 	"Node_ext_func",
@@ -444,6 +445,7 @@ flags2str(int flagval)
 		{ NUMCUR, "NUMCUR" },
 		{ NUMBER, "NUMBER" },
 		{ USER_INPUT, "USER_INPUT" },
+		{ BOOLVAL, "BOOL" },
 		{ INTLSTR, "INTLSTR" },
 		{ NUMINT, "NUMINT" },
 		{ INTIND, "INTIND" },
@@ -544,7 +546,6 @@ posix_compare(NODE *s1, NODE *s2)
 		s1->stptr[s1->stlen] = save1;
 		s2->stptr[s2->stlen] = save2;
 	}
-#if ! defined(__DJGPP__)
 	else {
 		/* Similar logic, using wide characters */
 		const wchar_t *p1, *p2;
@@ -577,7 +578,6 @@ posix_compare(NODE *s1, NODE *s2)
 			}
 		}
 	}
-#endif
 
 	return ret;
 }
@@ -1168,11 +1168,19 @@ r_get_lhs(NODE *n, bool reference)
 		n->var_value = dupnode(Nnull_string);
 		break;
 
+	case Node_elem_new:
+		efree(n->stptr);
+		n->stptr = NULL;
+		n->stlen = 0;
+		n->type = Node_var;
+		n->var_value = dupnode(Nnull_string);
+		break;
+
 	case Node_var:
 		break;
 
 	default:
-		cant_happen();
+		cant_happen("unexpected variable type %s", nodetype2str(n->type));
 	}
 
 	if (do_lint && reference && var_uninitialized(n))
@@ -1314,6 +1322,7 @@ setup_frame(INSTRUCTION *pc)
 		switch (m->type) {
 		case Node_var_new:
 		case Node_var_array:
+		case Node_elem_new:
 			r->type = Node_array_ref;
 			r->orig_array = r->prev_array = m;
 			break;
@@ -1346,7 +1355,7 @@ setup_frame(INSTRUCTION *pc)
 			break;
 
 		default:
-			cant_happen();
+			cant_happen("unexpected parameter type %s", nodetype2str(m->type));
 		}
 		r->vname = fp[i].param;
 	}
@@ -1521,31 +1530,113 @@ eval_condition(NODE *t)
 	return boolval(t);
 }
 
-typedef enum {
-	SCALAR_EQ_NEQ,
-	SCALAR_RELATIONAL
-} scalar_cmp_t;
+static bool cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type);
+extern bool mpg_cmp_as_numbers(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type);
 
 /* cmp_scalars -- compare two nodes on the stack */
 
-static inline int
+static bool
 cmp_scalars(scalar_cmp_t comparison_type)
 {
 	NODE *t1, *t2;
 	int di;
+	bool ret;
 
 	t2 = POP_SCALAR();
 	t1 = TOP();
+
+	t1 = elem_new_to_scalar(t1);
+	t2 = elem_new_to_scalar(t2);
+
 	if (t1->type == Node_var_array) {
 		DEREF(t2);
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t1));
 	}
-	di = cmp_nodes(t1, t2, comparison_type == SCALAR_EQ_NEQ);
+
+	if ((t1->flags & STRING) != 0 || (t2->flags & STRING) != 0) {
+		bool use_strcmp = (comparison_type == SCALAR_EQ || comparison_type == SCALAR_NEQ);
+		di = cmp_nodes(t1, t2, use_strcmp);
+
+		switch (comparison_type) {
+		case SCALAR_EQ:
+			ret = (di == 0);
+			break;
+		case SCALAR_NEQ:
+			ret = (di != 0);
+			break;
+		case SCALAR_LT:
+			ret = (di < 0);
+			break;
+		case SCALAR_LE:
+			ret = (di <= 0);
+			break;
+		case SCALAR_GT:
+			ret = (di > 0);
+			break;
+		case SCALAR_GE:
+			ret = (di >= 0);
+			break;
+		}
+	} else {
+		fixtype(t1);
+		fixtype(t2);
+
+#ifdef HAVE_MPFR
+		if (do_mpfr)
+			ret = mpg_cmp_as_numbers(t1, t2, comparison_type);
+		else
+#endif
+			ret = cmp_doubles(t1, t2, comparison_type);
+	}
+
 	DEREF(t1);
 	DEREF(t2);
-	return di;
+	return ret;
 }
 
+
+/* cmp_doubles --- compare two doubles */
+
+static bool
+cmp_doubles(const NODE *t1, const NODE *t2, scalar_cmp_t comparison_type)
+{
+	/*
+	 * This routine provides numeric comparisons that should work
+	 * the same as in C.  It should NOT be used for sorting.
+	 */
+
+	bool t1_nan = isnan(t1->numbr);
+	bool t2_nan = isnan(t2->numbr);
+	int ret;
+
+	if ((t1_nan || t2_nan) && comparison_type != SCALAR_NEQ)
+		return false;
+
+	switch (comparison_type) {
+	case SCALAR_EQ:
+		ret = (t1->numbr == t2->numbr);
+		break;
+	case SCALAR_NEQ:
+		ret = (t1->numbr != t2->numbr);
+		break;
+	case SCALAR_LT:
+		ret = (t1->numbr < t2->numbr);
+		break;
+	case SCALAR_LE:
+		ret = (t1->numbr <= t2->numbr);
+		break;
+	case SCALAR_GT:
+		ret = (t1->numbr > t2->numbr);
+		break;
+	case SCALAR_GE:
+		ret = (t1->numbr >= t2->numbr);
+		break;
+	}
+
+	return ret;
+}
+
+
 /* op_assign --- assignment operators excluding = */
 
 static void
@@ -1795,3 +1886,37 @@ init_interpret()
 		interpret = r_interpret;
 }
 
+/* elem_new_to_scalar --- convert Node_elem_new to untyped scalar */
+
+NODE *
+elem_new_to_scalar(NODE *n)
+{
+	if (n->type != Node_elem_new)
+		return n;
+
+	if (n->valref > 1) {
+		unref(n);
+		return dupnode(Nnull_string);
+	}
+
+	n->type = Node_val;
+
+	return n;
+}
+
+/* fix_nan_sign --- fix NaN sign on RiscV */
+
+// See the thread starting at
+// https://lists.gnu.org/archive/html/bug-gawk/2022-09/msg00005.html
+// for why we need this function.
+
+static double
+fix_nan_sign(double left, double right, double result)
+{
+	if (isnan(left) && signbit(left))
+		return copysign(result, -1.0);
+	else if (isnan(right) && signbit(right))
+		return copysign(result, -1.0);
+	else
+		return result;
+}
