@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2020 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2022 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -81,15 +81,17 @@ static awk_fieldwidth_info_t *FIELDWIDTHS = NULL;
 
 NODE **fields_arr;		/* array of pointers to the field nodes */
 bool field0_valid;		/* $(>0) has not been changed yet */
-int default_FS;			/* true when FS == " " */
-Regexp *FS_re_yes_case = NULL;
-Regexp *FS_re_no_case = NULL;
-Regexp *FS_regexp = NULL;
-Regexp *FPAT_re_yes_case = NULL;
-Regexp *FPAT_re_no_case = NULL;
-Regexp *FPAT_regexp = NULL;
+static int default_FS;		/* true when FS == " " */
+static Regexp *FS_re_yes_case = NULL;
+static Regexp *FS_re_no_case = NULL;
+static Regexp *FS_regexp = NULL;
+static Regexp *FPAT_re_yes_case = NULL;
+static Regexp *FPAT_re_no_case = NULL;
+static Regexp *FPAT_regexp = NULL;
 NODE *Null_field = NULL;
 
+#define clear_mpfr(n) ((n)->flags &= ~(MPFN | MPZN | NUMCUR))
+
 /* init_fields --- set up the fields array to start with */
 
 void
@@ -230,6 +232,7 @@ rebuild_record()
 			}
 
 			n->stptr = cops;
+			clear_mpfr(n);
 			unref(r);
 			fields_arr[i] = n;
 			assert((n->flags & WSTRCUR) == 0);
@@ -347,6 +350,11 @@ reset_record()
 	}
 }
 
+/*
+ * purge_record --- throw away the fields, make sure that
+ * 	individual nodes remain valid.
+ */
+
 static void
 purge_record()
 {
@@ -829,7 +837,7 @@ fw_parse_field(long up_to,	/* parse only up to this field number */
 
 /* invalidate_field0 --- $0 needs reconstruction */
 
-void
+static void
 invalidate_field0()
 {
 	field0_valid = false;
@@ -971,6 +979,8 @@ do_split(int nargs)
 			 Regexp *, Setfunc, NODE *, NODE *, bool);
 	Regexp *rp = NULL;
 
+	check_args_min_max(nargs, "split", 3, 4);
+
 	if (nargs == 4) {
 		static bool warned = false;
 
@@ -980,6 +990,8 @@ do_split(int nargs)
 		sep_arr = POP_PARAM();
 		if (sep_arr->type != Node_var_array)
 			fatal(_("split: fourth argument is not an array"));
+		check_symtab_functab(sep_arr, "split",
+				_("%s: cannot use %s as fourth argument"));
 		if ((do_lint_extensions || do_lint_old) && ! warned) {
 			warned = true;
 			lintwarn(_("split: fourth argument is a gawk extension"));
@@ -990,6 +1002,8 @@ do_split(int nargs)
 	arr = POP_PARAM();
 	if (arr->type != Node_var_array)
 		fatal(_("split: second argument is not an array"));
+	check_symtab_functab(arr, "split",
+			_("%s: cannot use %s as second argument"));
 
 	if (sep_arr != NULL) {
 		if (sep_arr == arr)
@@ -1069,15 +1083,21 @@ do_patsplit(int nargs)
 	char *s;
 	Regexp *rp = NULL;
 
+	check_args_min_max(nargs, "patsplit", 3, 4);
+
 	if (nargs == 4) {
 		sep_arr = POP_PARAM();
 		if (sep_arr->type != Node_var_array)
 			fatal(_("patsplit: fourth argument is not an array"));
+		check_symtab_functab(sep_arr, "patsplit",
+				_("%s: cannot use %s as fourth argument"));
 	}
 	sep = POP();
 	arr = POP_PARAM();
 	if (arr->type != Node_var_array)
 		fatal(_("patsplit: second argument is not an array"));
+	check_symtab_functab(arr, "patsplit",
+			_("%s: cannot use %s as second argument"));
 
 	src = TOP_STRING();
 
