@@ -2,7 +2,7 @@
 ** popen.h -- prototypes for pipe functions
 */
 /*
- * Copyright (C) 2010, 2013, 2014, 2016, the Free Software Foundation, Inc.
+ * Copyright (C) 2010, 2013, 2014, 2016, 2022, the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -22,7 +22,6 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
  */
 
-#if !defined (__DJGPP__)
 # if defined (popen)
 #  undef popen
 #  undef pclose
@@ -31,11 +30,10 @@
 # define pclose(f)	os_pclose(f)
   extern FILE *os_popen( const char *, const char * );
   extern int  os_pclose( FILE * );
-# ifdef __MINGW32__
+#ifdef __MINGW32__
 #  define system(c)	os_system(c)
    extern int os_system( const char * );
 #  define SIGKILL	9
    extern int kill( int, int );
    extern char *quote_cmd( const char * );
-# endif	 /* __MINGW32__ */
-#endif	/* !__DJGPP__ */
+#endif	 /* __MINGW32__ */
